/*
	cdrom.c	by	BERO
	info from psx.rules.org / doomed@c64.org
*/

#include <stdio.h>
#include <string.h>
#include <conio.h>

#include "fpse.h"

#define	RAMADDR(addr)     ((void*)&ram[(addr)&0x1fffff])

#define USE_EUROPE_PSX    1
#define USE_AMERICAN_PSX  2
#define USE_JAPAN_PSX     3

// #define PSX_COUNTRY       USE_EUROPE_PSX
#define PSX_COUNTRY       USE_AMERICAN_PSX
// #define PSX_COUNTRY       USE_JAPAN_PSX

enum {
    CdlSync,    CdlNop,       CdlSetloc,  CdlPlay,      // 00-03
    CdlForward, CdlBackword,  CdlReadN,   CdlStandby,   // 04-07
    CdlStop,    CdlPause,     CdlInit,    CdlMute,      // 08-11
    CdlDemute,  CdlSetfilter, CdlSetmode, CdlGetparam,  // 12-15
    CdlGetlocL, CdlGetlocP,   Cdl18,      CdlGetTN,     // 16-19
    CdlGetTD,   CdlSeekL,     CdlSeekP,   Cdl23,        // 20-23
    Cdl24,      CdlTest,      CdlCheckID, CdlReadS,     // 24-27
    CdlReset                                           // 25
};

enum {
	ModeSpeed     = 0x80,        /* 0: normal speed  1: double speed */
	ModeRT        = 0x40,        /* 0: ADPCM off     1: ADPCM on     */
	ModeSize1     = 0x20,        /* 0: 2048 byte     1: 2340 byte    */
	ModeSize0     = 0x10,        /* 0: -             1: 2328 byte    */
	ModeSF        = 0x08,        /* 0: Channel off   1: Channel on   */
	ModeRept      = 0x04,        /* 0: Report off    1: Report on    */
	ModeAP        = 0x02,        /* 0: AutoPause off 1: AutoPause on */
	ModeDA        = 0x01         /* 0: CD-DA off     1: CD-DA on     */
};

enum {
	StatPlay      = 0x80,        /* playing CD-DA */
	StatSeek      = 0x40,        /* seeking */
	StatRead      = 0x20,        /* reading data frames */
	StatShellOpen = 0x10,        /* once shell open */
	StatSeekError = 0x04,        /* seek error detected */
	StatStandby   = 0x02,        /* spindle motor rotating */
	StatError     = 0x01         /* command error detected */
};

enum {
	NoIntr      = 0x00,        /* No interrupt */
	DataReady   = 0x01,        /* Data Ready */
	Acknowledge = 0x02,        /* Acknowledge */
	Complete    = 0x03,        /* Command Complete */
	DataEnd     = 0x04,        /* End of Data Detected */
	DiskError   = 0x05        /* Error Detected */
};

static struct {
	UINT8 stat;
	UINT8 ctrl;
	UINT8 mode;
        UINT8 Status;

	int reset;

	UINT8 param[8];
	int paramptr;
	/* param */
	UINT8 result[8];
	int resultptr;
	int resultsize;
	int resultready;
	struct {
		UINT8 minute;
		UINT8 second;
		UINT8 frame;
                UINT8 pad;
	} loc;
	UINT8	file;
	UINT8	chan;
// Async Related
        UINT8   NextStat;
        UINT8   DmaReady;
        UINT32  NextSize;
// Buffers
        UINT8  *SectorBuf;
        UINT32  SectorSize;
} cd;

#if PSX_COUNTRY == USE_EUROPE_PSX
static UINT8 Cmd25P22[8]={ 0x66,0x6F,0x72,0x20,0x55,0x2F,0x43,0x00 }; // "for U/C"
static UINT8 Cmd25P23[8]={ 0x43,0x58,0x44,0x32,0x35,0x34,0x35,0x51 }; // "CXD2545Q"
static UINT8 Cmd25P24[8]={ 0x43,0x58,0x44,0x32,0x35,0x34,0x35,0x51 }; // "CXD2545Q"
static UINT8 Cmd25P25[8]={ 0x43,0x58,0x44,0x31,0x38,0x31,0x35,0x51 }; // "CXD1815Q"
static UINT8 SCEx[6] = { 0x09, 0xA9, 0x3D, 0x2B, 0xA5, 0x74 };  // SCEE European
#endif

#if PSX_COUNTRY == USE_AMERICAN_PSX
static UINT8 Cmd25P22[8]={ 0x66,0x6F,0x72,0x20,0x55,0x53,0x2F,0x41 }; // "for US/A"
static UINT8 Cmd25P23[8]={ 0x43,0x58,0x44,0x31,0x37,0x38,0x32,0x42 }; // "CXD1782B"
static UINT8 Cmd25P24[8]={ 0x43,0x58,0x44,0x32,0x35,0x31,0x36,0x51 }; // "CXD2516Q"
static UINT8 Cmd25P25[8]={ 0x43,0x58,0x44,0x31,0x31,0x39,0x39,0x42 }; // "CXD1199B"
static UINT8 SCEx[6] = { 0x09, 0xA9, 0x3D, 0x2B, 0xA5, 0xF4 };  // SCEA U.S.
#endif

#if PSX_COUNTRY == USE_JAPAN_PSX
static UINT8 Cmd25P22[8]={ 0x66,0x6F,0x72,0x20,0x4A,0x61,0x70,0x61 }; // "for Japa"
static UINT8 Cmd25P23[8]={ 0x43,0x58,0x44,0x32,0x35,0x34,0x35,0x51 }; // "CXD2545Q"
static UINT8 Cmd25P24[8]={ 0x43,0x58,0x44,0x32,0x35,0x34,0x35,0x51 }; // "CXD2545Q"
static UINT8 Cmd25P25[8]={ 0x43,0x58,0x44,0x31,0x38,0x31,0x35,0x51 }; // "CXD1815Q"
static UINT8 SCEx[6] = { 0x09, 0xA9, 0x3D, 0x2B, 0xA5, 0x74 };  // SCEI Japanese
#endif

UINT8 *cd_country_idx[4] = { Cmd25P22,Cmd25P23,Cmd25P24,Cmd25P25 };

/*
	result[0]

	0x80 CD-DA Play
	0x02 working?

*/

static int get_sector_size()
{
    switch( cd.mode & 0x30 ) {
        case 0x00: return 512*4;
        case 0x10: return 582*4;
        case 0x20: return 585*4;
    }
    return 0;
}

static void cdrom_async()
{
    cd.stat         = cd.NextStat;
    cd.resultsize   = cd.NextSize;
    cd.resultptr    = 0;
    cd.resultready  = 1;
    cd.ctrl &= ~0x80;
    cd.ctrl |= cd.DmaReady;
    cd.DmaReady = 0;
}

void cd_initvar()
{
 memset(&cd,0,sizeof(cd));
 cd.Status = 0x2;
 EventCallBack[INT_CD] = cdrom_async;
}

int cd0_read(void)
{
	int ret = cd.ctrl;
	if (cd.resultsize && cd.resultready) ret|=0x20; else ret&=~0x20;
//        printf("read cd0 %x\n",ret|0x08);
	return ret|0x08;
}

int cd1_read(void)
{
	int ret=0;
	if (cd.resultsize) {
	  ret = cd.result[cd.resultptr++];
	  if (cd.resultptr>=cd.resultsize) {
	    cd.resultptr = cd.resultsize = 0;
		cd.resultready = 0;
	  }
	}
        printf("read cd1 %x\n",ret);
	return ret;
}

int cd2_read(void)
{
//        printf("read cd2 ,0\n");
	return 0;
}

int cd3_read(void)
{
//        printf("read cd3 %x\n",cd.stat);
	return cd.stat;
}

#define	RESULT(n)       cd.result[0] = n; cd.resultsize = 1
#define	SRESULT         cd.result[0] = cd.Status; cd.resultsize = 1

#define	INT2BCD(a)	((a/10)*16+(a%10))
#define	BCD2INT(a)	((a>>4)*10+(a&0xf))

void cd0_write(int data)
{
/*	write 0??
	write read & 3?
*/
//        printf("write cd0 %x\n",data);
	cd.ctrl = data;
	if (data==0) {
		cd.paramptr = 0;
		cd.resultready = 0;
	}
	if (data&1) cd.resultready = 1;
	if (data==1) cd.reset = 1;
}

void cd2_write(int data)
{
//        printf("write cd2 %x\n",data);
	if (cd.reset==2 && data==7) {
		cd.paramptr = 0;
		cd.resultready = 1;
		cd.reset = 0;
		cd.stat = 0;
		cd.ctrl = 0;
		return;
	}
	cd.reset = 0;
	if (cd.paramptr<8)
		cd.param[cd.paramptr++] = data;
}

void cd3_write(int data)
{
//	cd.stat = data;
//        printf("write cd3 %x\n",data);
	if (data==7 && cd.reset==1) cd.reset=2;
	else cd.reset=0;
	if (cd.resultready && data==7) {
		cd.resultsize = cd.resultptr = 0;
		cd.stat = 0;
	}
}

void cd1_write(int data)
{
    static char *cmdname[]= {
    "CdlSync",    "CdlNop",       "CdlSetloc",  "CdlPlay",
    "CdlForward", "CdlBackword",  "CdlReadN",   "CdlStandby",
    "CdlStop",    "CdlPause",     "CdlReset",   "CdlMute",
    "CdlDemute",  "CdlSetfilter", "CdlSetmode", "CdlGetparam",
    "CdlGetlocL", "CdlGetlocP",   "Cdl18",      "CdlGetTN",
    "CdlGetTD",   "CdlSeekL",     "CdlSeekP",   "Cdl23",
    "Cdl24",      "CdlTest",      "CdlCheckID", "CdlReadS"
    };

    if (data<28) {
        PRINTF("%s\n",cmdname[data]);
    }

printf("Command %x - Code %02x %02x %02x %02x %02x %02x %02x %02x\n",
               data,
               cd.param[0], cd.param[1],
               cd.param[2], cd.param[3],
               cd.param[4], cd.param[5],
               cd.param[6], cd.param[7]);

    switch(data){
    case CdlNop:
        cd.stat = 3;
        SRESULT;
        break;

    case CdlSetloc:
        cd.stat = 2;
        cd.loc.minute = cd.param[0];
        cd.loc.second = cd.param[1];
        cd.loc.frame = cd.param[2];
        SRESULT;
        break;

    case CdlPlay:
        cd.stat = 2;
        cd_play((UINT8 *)&cd.loc);
        cd.Status |= StatPlay;
        SRESULT;

        Event_Register |= 0x80000000 | (1<<INT_CD);
        Event_List[INT_CD] = 0x00000002;
        cd.NextStat = 3;
        cd.NextSize = 1;
        cd.ctrl |= 0x80;
        return;

    case CdlForward:
    case CdlBackword:
        cd.stat = 2;
        SRESULT;
        break;

    case CdlReadN:
        cd.SectorBuf  = cd_read((UINT8 *)&cd.loc);
        cd.SectorSize = get_sector_size();
        cd.Status |= 2;
        SRESULT;
        cd.stat = 0;
        cd.resultsize = 1;
        if (++cd.loc.frame >= 75) {
            cd.loc.frame = 0;
            if (++cd.loc.second >= 60) {
                cd.loc.second = 0;
                if (++cd.loc.minute >= 74)
                    cd.loc.minute = 0;
            }
        }
        Event_Register |= 0x80000000 | (1<<INT_CD);
        Event_List[INT_CD] = 0x00000003;
        cd.NextStat = 1;
        cd.NextSize = 1;
        cd.DmaReady = 0x40;
        cd.ctrl |= 0x80;
        return;

    case CdlInit:
        cd.mode = 0;
    case CdlReset:
    case CdlStandby:
        cd.stat = 2;
        cd.Status = (cd.Status & ~StatPlay) | StatStandby;
        SRESULT;
        break;

    case CdlStop:
        cd.stat = 2;
        cd.Status &= ~(StatPlay | StatStandby);
        cd_stop();
        SRESULT;
        break;

    case CdlPause:
        cd.stat = 0;
        if (cd.mode & ModeDA) {
            cd.Status = (cd.Status ^ StatPlay) | StatStandby;
            if (cd.Status & StatPlay)
                cd_play((unsigned char *)&cd.loc);
        } else
            cd.Status = (cd.Status & ~StatPlay) | StatStandby;
        SRESULT;

        Event_Register |= 0x80000000 | (1<<INT_CD);
        Event_List[INT_CD] = 0x00000003;
        cd.NextStat = 2;
        cd.NextSize = 1;
        cd.ctrl |= 0x80;
        return;

    case CdlMute:
    case CdlDemute:
        cd.stat = 2;
        SRESULT;
        break;

    case CdlSetfilter:
        cd.stat = 2;
        cd.file = cd.param[0];
        cd.chan = cd.param[1];
        SRESULT;
        break;

    case CdlSetmode:
        cd.stat = 3;
        cd.mode = cd.param[0];
        SRESULT;
        break;

    case CdlGetparam:
        cd.stat = 2;
        cd.result[0] = cd.Status; /* status */
        cd.result[1] = cd.mode;
        cd.result[2] = cd.file; /* file */
        cd.result[3] = cd.chan; /* channel */
        cd.result[4] = 0;
        cd.result[5] = 0;
        cd.resultsize = 6;
        break;

    case CdlGetlocL:
        cd.stat = 2;
        cd.result[0] = cd.loc.minute; /* min */
        cd.result[1] = cd.loc.second; /* sec */
        cd.result[2] = cd.loc.frame; /* frame  */
        cd.result[3] = cd.mode; /* mode */
        cd.result[4] = cd.file; /* file */
        cd.result[5] = cd.chan; /* channel */
        cd.resultsize = 6;
        break;

    case CdlGetlocP:
        cd.stat = 2;
        cd.result[0] = 0; /* track */
        cd.result[1] = 0; /* index */
        cd.result[2] = 0; /* min  */
        cd.result[3] = 0; /* sec */
        cd.result[4] = 0; /* frame */
        cd.result[5] = 0; /* min  */
        cd.result[6] = 0; /* sec */
        cd.result[7] = 0; /* frame */
        cd.resultsize = 8;
        break;

    case CdlGetTN:
        cd.stat = 2;
        cd_GetTN(cd.result);
        cd.result[0] = cd.Status; /* status */
        cd.result[1] = INT2BCD(cd.result[1]); /* first track */
        cd.result[2] = INT2BCD(cd.result[2]); /* total track */
        cd.resultsize = 3;
        break;

    case CdlGetTD:
        cd.stat = 2;
        cd_GetTD(cd.result,cd.param[0]);
	cd.result[0] = cd.Status; /* status */
	cd.result[1] = INT2BCD(cd.result[1]); /* min */
	cd.result[2] = INT2BCD(cd.result[2]); /* sec */
        cd.resultsize = 3;
        break;

    case CdlSeekL:
    case CdlSeekP:
        cd.stat = 2;
        SRESULT;
        break;

    case CdlTest:
        switch (cd.param[0]) {
        case 0x20:
        case 0x22:
            memcpy(cd.result,Cmd25P22,8);
            break;
        case 0x23:
            memcpy(cd.result,Cmd25P23,8);
            break;
        case 0x24:
            memcpy(cd.result,Cmd25P24,8);
            break;
        case 0x25:
            memcpy(cd.result,Cmd25P25,8);
            break;
        }
        cd.resultsize = 8;
        cd.stat = 3;
        break;

    case CdlCheckID:
        cd.result[0] = 0x22;
        cd.result[1] = 0;
        cd.result[2] = 0x22;
        cd.result[3] = 0;
        cd.result[4] = 'S';
        cd.result[5] = 'C';
        cd.result[6] = 'E';
        cd.result[7] = 'I';

        cd.resultsize = 8;
        cd.stat = 2;
        break;

    case CdlReadS:
        cd_read((unsigned char *)&cd.loc);
        SRESULT;
        cd.stat = 2;
        break;

    default:
        printf("unknown cd command %d - %xh\n",data,data);
        cd.stat = 2;
    }
    cd.resultptr = 0;
    cd.resultready = 1;

/* cd_interrupt happen */
    Irq_Pulse |= (1<<INT_CD);
}

void dma3_exec(UINT32 adr,UINT32 bcr,UINT32 chcr)
{
    int size;

    printf("cdma called: %08x %08x %08x\n",(int)adr,(int)bcr,(int)chcr);

// get the size in bytes
    size = (bcr>>16)*(bcr&0xffff)*4;

// check the dma_chcr
    switch(chcr) {
    case 0x11000000:
        memcpy(RAMADDR(adr),cd.SectorBuf+2340-cd.SectorSize,size);
        break;
    }
// Clear buffer data ready
    cd.ctrl &= ~0x40;
}
